import unittest

import plotly.graph_objects as go

from random_events.interval import *
from random_events.product_algebra import SimpleEvent, Event
from random_events.set import SetElement, Set
from random_events.sigma_algebra import AbstractSimpleSet
from random_events.variable import Continuous, Symbolic


str_set = {'a', 'c', 'b'}

class EventTestCase(unittest.TestCase):
    x = Continuous("x")
    y = Continuous("y")
    z = Continuous("z")
    a = Symbolic("a", str_set)
    b = Symbolic("b", str_set)

    def test_constructor(self):
        sa = SetElement("a", str_set)
        sb = SetElement("b", str_set)
        event = SimpleEvent({self.a: Set(sa), self.x: Interval(SimpleInterval(0, 1)), self.y: Interval(SimpleInterval(0, 1))})

        self.assertEqual(event[self.x], Interval(SimpleInterval(0, 1)))
        self.assertEqual(event[self.y], Interval(SimpleInterval(0, 1)))
        self.assertEqual(event[self.a], Set(sa))

        self.assertFalse(event.is_empty())

    def test_intersection_with(self):
        sa = SetElement("a", str_set)
        sb = SetElement("b", str_set)
        sc = SetElement("c", str_set)
        event_1 = SimpleEvent(
            {self.a: Set(sa, sb), self.x: Interval(SimpleInterval(0, 1)), self.y: Interval(SimpleInterval(0, 1))})
        event_2 = SimpleEvent({self.a: Set(sa), self.x: Interval(SimpleInterval(0.5, 1))})
        event_3 = SimpleEvent({self.a: Set(sc)})
        intersection = event_1.intersection_with(event_2)

        intersection_ = SimpleEvent(
            {self.a: Set(sa), self.x: Interval(SimpleInterval(0.5, 1)), self.y: Interval(SimpleInterval(0, 1))})

        self.assertEqual(intersection, intersection_)
        self.assertNotEqual(intersection, event_1)

        second_intersection = event_1.intersection_with(event_3)
        self.assertTrue(second_intersection.is_empty())

    def test_complement(self):
        sa = SetElement("a", str_set)
        sb = SetElement("b", str_set)
        sc = SetElement("c", str_set)
        event = SimpleEvent({self.a: Set(sa, sb), self.x: Interval(SimpleInterval(0, 1)), self.y: self.y.domain})
        complement = event.complement()
        self.assertEqual(len(complement), 2)
        complement_1 = SimpleEvent({self.a: Set(sc), self.x: self.x.domain, self.y: self.y.domain})
        complement_2 = SimpleEvent({self.a: event[self.a], self.x: event[self.x].complement(), self.y: self.y.domain})
        self.assertEqual(complement, SortedSet([complement_1, complement_2]))

    def test_simplify(self):
        sa = SetElement("a", str_set)
        sb = SetElement("b", str_set)
        sc = SetElement("c", str_set)
        event_1 = SimpleEvent(
            {self.a: Set(sa, sb), self.x: Interval(SimpleInterval(0, 1)), self.y: Interval(SimpleInterval(0, 1))})
        event_2 = SimpleEvent(
            {self.a: Set(sc), self.x: Interval(SimpleInterval(0, 1)), self.y: Interval(SimpleInterval(0, 1))})
        event = Event(event_1, event_2)
        simplified = event.simplify()
        self.assertEqual(len(simplified.simple_sets), 1)

        result = Event(SimpleEvent(
            {self.a: self.a.domain, self.x: Interval(SimpleInterval(0, 1)), self.y: Interval(SimpleInterval(0, 1))}))
        self.assertEqual(simplified, result)

    def test_simplify_once(self):
        sa = SetElement("a", str_set)
        sb = SetElement("b", str_set)
        sc = SetElement("c", str_set)
        event_1 = SimpleEvent(
            {self.a: Set(sa, sb), self.x: Interval(SimpleInterval(0, 1)), self.y: Interval(SimpleInterval(0, 1))})
        event_2 = SimpleEvent(
            {self.a: Set(sc), self.x: Interval(SimpleInterval(0, 1)), self.y: Interval(SimpleInterval(0, 1))})
        event = Event(event_1, event_2)
        simplified = event.simplify_once()
        self.assertEqual(len(simplified[0].simple_sets), 1)

        result = Event(SimpleEvent(
            {self.a: self.a.domain, self.x: Interval(SimpleInterval(0, 1)), self.y: Interval(SimpleInterval(0, 1))}))
        self.assertEqual(simplified[0], result)

    def test_to_json(self):
        sa = SetElement("a", str_set)
        sb = SetElement("b", str_set)
        sc = SetElement("c", str_set)
        event = SimpleEvent({self.a: Set(sa, sb), self.x: Interval(SimpleInterval(0, 1)),
                             self.y: Interval(SimpleInterval(0, 1))})
        event_ = AbstractSimpleSet.from_json(event.to_json())
        self.assertEqual(event_, event)

    def test_plot_2d(self):
        event_1 = SimpleEvent({self.x: Interval(SimpleInterval(0, 1)), self.y: Interval(SimpleInterval(0, 1))})
        event_2 = SimpleEvent({self.x: Interval(SimpleInterval(1, 2)), self.y: Interval(SimpleInterval(1, 2))})
        event = Event(event_1, event_2)
        fig = go.Figure(event.plot(), event.plotly_layout())
        self.assertIsNotNone(fig)
        # fig.show()

    def test_plot_3d(self):
        event_1 = SimpleEvent(
            {self.x: SimpleInterval(0, 1), self.y: SimpleInterval(0, 1), self.z: SimpleInterval(0, 1)})
        event_2 = SimpleEvent(
            {self.x: SimpleInterval(1, 2), self.y: SimpleInterval(1, 2), self.z: SimpleInterval(1, 2)})
        event = Event(event_1, event_2)
        fig = go.Figure(event.plot(), event.plotly_layout())
        self.assertIsNotNone(fig)
        # fig.show()

    def test_union(self):
        sa = SetElement("a", str_set)
        sb = SetElement("b", str_set)
        sc = SetElement("c", str_set)
        event = Event(SimpleEvent({self.a: sa, self.x: open(-float("inf"), 2)}))
        second_event = SimpleEvent({self.a: Set(sa, sb), self.x: open(1, 4)}).as_composite_set()
        union = event | second_event
        result = Event(SimpleEvent({self.a: sa, self.x: open(-float("inf"), 4)}),
                        SimpleEvent({self.a: sb, self.x: open(1, 4)}))
        self.assertEqual(union, result)

    def test_marginal_event(self):
        event_1 = SimpleEvent({self.x: closed(0, 1), self.y: Interval(SimpleInterval(0, 1))})
        event_2 = SimpleEvent({self.x: closed(1, 2), self.y: Interval(SimpleInterval(3, 4))})
        event_3 = SimpleEvent({self.x: closed(5, 6), self.y: Interval(SimpleInterval(5, 6))})
        event = Event(event_1, event_2, event_3)
        marginal = event.marginal(SortedSet([self.x]))
        self.assertEqual(marginal, SimpleEvent({self.x: closed(0, 2) | closed(5, 6)}).as_composite_set())
        fig = go.Figure(marginal.plot())
        # fig.show()

    def test_to_json_multiple_events(self):
        sa = SetElement("a", str_set)
        sb = SetElement("b", str_set)
        sc = SetElement("c", str_set)
        event = SimpleEvent({self.x: closed(0, 1), self.y: SimpleInterval(3, 5),
                             self.a: Set(sa, sb)}).as_composite_set()
        event_ = AbstractSimpleSet.from_json(event.to_json())
        self.assertEqual(event_, event)

    def test_bounding_box(self):
        event_1 = SimpleEvent({self.x: closed(0, 1), self.y: SimpleInterval(0, 1)}).as_composite_set()
        event_2 = SimpleEvent({self.x: closed(1, 2), self.y: Interval(SimpleInterval(3, 4))}).as_composite_set()
        event = event_1 | event_2

        event_before_bounding_box = event.__deepcopy__()
        bounding_box = event.bounding_box()
        result = SimpleEvent({self.x: closed(0, 2),
                              self.y: SimpleInterval(0, 1).as_composite_set() | SimpleInterval(3, 4).as_composite_set()})
        self.assertEqual(bounding_box, result)
        self.assertEqual(event, event_before_bounding_box)

    def test_complex_event_bounding_box_with_references(self):
        event1 = SimpleEvent({self.x: closed(0, 1) |  closed(2, 3),
                             self.y: closed(0, 1) |  closed(2, 3)}).as_composite_set()
        event2 = SimpleEvent({self.x: closed(1, 2) |  closed(3, 4),
                             self.y: closed(1, 2) |  closed(3, 4)}).as_composite_set()
        event = event1 | event2
        event_before_bb = event.__deepcopy__()
        bb = event.bounding_box()
        self.assertEqual(event, event_before_bb)

class NoneTypeObjectInDifferenceTestCase(unittest.TestCase):
    x: Continuous = Continuous("x")
    y: Continuous = Continuous("y")
    event_1 = Event(SimpleEvent({x: SimpleInterval(0, 0.25, Bound.CLOSED, Bound.CLOSED),
                                 y: SimpleInterval(0.25, 1, Bound.CLOSED, Bound.CLOSED)}),
                    SimpleEvent({x: SimpleInterval(0.75, 1, Bound.CLOSED, Bound.CLOSED),
                                 y: SimpleInterval(0.75, 1, Bound.CLOSED, Bound.CLOSED)}))
    event_2 = SimpleEvent({x: SimpleInterval(0., 0.25, Bound.CLOSED, Bound.CLOSED),
                           y: SimpleInterval(0., 1, Bound.CLOSED, Bound.CLOSED)}).as_composite_set()

    def test_union(self):
        union = self.event_1 | self.event_2
        union_by_hand = Event(SimpleEvent({self.x: SimpleInterval(0, 0.25, Bound.CLOSED, Bound.CLOSED),
                                           self.y: SimpleInterval(0., 1, Bound.CLOSED, Bound.CLOSED)}),
                              SimpleEvent({self.x: SimpleInterval(0.75, 1, Bound.CLOSED, Bound.CLOSED),
                                           self.y: SimpleInterval(0.75, 1, Bound.CLOSED, Bound.CLOSED)}))
        self.assertEqual(union, union_by_hand)


class OperationsWithEmptySetsTestCase(unittest.TestCase):
    sa = SetElement("a", str_set)
    sb = SetElement("b", str_set)
    sc = SetElement("c", str_set)
    x: Continuous = Continuous("x")
    y: Continuous = Continuous("y")

    def test_empty_union(self):
        empty_event = SimpleEvent({self.x: SimpleInterval(0, 0), self.y: SimpleInterval(0, 0)}).as_composite_set()
        event = SimpleEvent({self.x: SimpleInterval(0, 1), self.y: SimpleInterval(0, 1)}).as_composite_set()
        union = empty_event.union_with(event)
        self.assertEqual(union, event)

    def test_union_different_variables(self):
        event_1 = SimpleEvent({self.x: SimpleInterval(0, 1)}).as_composite_set()
        event_2 = SimpleEvent({self.y: SimpleInterval(0, 1)}).as_composite_set()
        union = event_1.union_with(event_2)
        # self.assertEqual(union, Event(event_1, event_2))

    def test_difference_with_empty_set(self):
        event = SimpleEvent({self.x: SimpleInterval(0, 1), self.y: SimpleInterval(0, 1)}).as_composite_set()
        empty_event = Event()
        diff = event.difference_with(empty_event)
        self.assertEqual(diff, event)

if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

from typing_extensions import Union

from .sigma_algebra import *
import random_events_lib as rl


class SetElement(AbstractSimpleSet):
    """
    Represents a SetElement.
    """

    def __init__(self, element, all_elements: Union[Set, SortedSet]):
        """
        Create a new set element.
        :param element: The element of the set. Can be a string, integer, float, or double.
        :param all_elements: The set of all elements.
        """
        self.all_elements = SortedSet(all_elements)
        self.hash_map = {hash(elem): elem for elem in self.all_elements}
        if element == EMPTY_SET_SYMBOL or element == -1:
            self.element = EMPTY_SET_SYMBOL
            self._cpp_object = rl.SetElement(set(self.hash_map.keys()))
        elif element not in self.all_elements:
            raise ValueError(f"Element {element} is not in the set of all elements.")
        else:
            self.element = element
            self.element_index = self.all_elements.index(element)
            self._cpp_object = rl.SetElement(self.element_index, set(self.hash_map.keys()))

    def _from_cpp(self, cpp_object):
        if cpp_object.element_index == -1:
            return SetElement(-1, set())
        return SetElement(list(self.hash_map.values())[cpp_object.element_index], self.all_elements)

    def contains(self, item: Self) -> bool:
        return self == item

    def non_empty_to_string(self) -> str:
        return self.element

    def __hash__(self):
        return hash((self.element, tuple(self.all_elements)))

    def __lt__(self, other):
        return self.element < other.element

    def __eq__(self, other):
        if self.element == other.element:
            return True
        return False

    def __repr__(self):
        return self.non_empty_to_string()

    def __iter__(self):
        return iter(self.all_elements)

    def to_json(self) -> Dict[str, Any]:
        return {**super().to_json(), "value": self.element, "content": self.all_elements}

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        return cls(data["value"], data["content"])

    def as_composite_set(self) -> AbstractCompositeSet:
        return Set(self)


class Set(AbstractCompositeSet):
    """
    Represents a set.
    """

    simple_sets: SetElementContainer
    """
    The simple sets that make up the set.
    """

    def __init__(self, *simple_sets):
        """
        Create a new set.
        :param simple_sets: The simple sets that make up the set.
        """
        super().__init__(*simple_sets)
        if len(simple_sets) > 0:
            self._cpp_object = rl.Set({simple_set._cpp_object for simple_set in self.simple_sets},
                                      self.simple_sets[0]._cpp_object.all_elements)
        else:
            self._cpp_object = rl.Set(set(), set())


    def _from_cpp(self, cpp_object):
        return Set(*[self.simple_sets[0]._from_cpp(cpp_simple_set) for cpp_simple_set in cpp_object.simple_sets])

    def complement_if_empty(self) -> Self:
        raise NotImplementedError("I don't know how to do this yet.")

    def new_empty_set(self) -> Self:
        return Set()


# Type definitions
if TYPE_CHECKING:
    SetElementContainer = SortedSet[SetElement]
else:
    SetElementContainer = SortedSet



from __future__ import annotations
import enum
import math
from typing import Dict, Any
from sortedcontainers import SortedSet
from typing_extensions import Self
from . import sigma_algebra
import random_events_lib as rl


class Bound(enum.Enum):
    """
    Enumerates the possible bounds for an interval.
    """

    CLOSED = 0
    """
    Represents a closed bound, i. e. the element is included from the interval.
    """

    OPEN = 1
    """
    Represents an open bound, i. e. the element is excluded in the interval.
    """

    def invert(self):
        return Bound.CLOSED if self == Bound.OPEN else Bound.OPEN

    def intersect(self, other: Self) -> Self:
        """
        Intersect with another border

        :param other: The other border
        :return: The intersection of the two borders
        """
        return Bound.OPEN if self == Bound.OPEN or other == Bound.OPEN else Bound.CLOSED


class SimpleInterval(sigma_algebra.AbstractSimpleSet):
    """
    Represents a simple interval.
    """

    lower: float = 0
    """
    The lower bound of the interval.
    """

    upper: float = 0
    """
    The upper bound of the interval.
    """

    left: Bound = Bound.OPEN
    """
    The bound type of the lower bound.
    """

    right: Bound = Bound.OPEN
    """
    The bound type of the upper bound.
    """

    def __init__(self, lower: float = 0, upper: float = 0, left: Bound = Bound.OPEN, right: Bound = Bound.OPEN):
        """
        Creates a new simple interval.
        :param lower: The lower bound of the interval.
        :param upper: The upper bound of the interval.
        :param left: The bound type of the lower bound.
        :param right: The bound type of the upper bound.
        """
        self.lower = lower
        self.upper = upper
        self.left = left
        self.right = right

        self._cpp_object = rl.SimpleInterval(lower, upper, left.value, right.value)

    @classmethod
    def _from_cpp(cls, cpp_object: rl.SimpleInterval) -> Self:
        return cls(cpp_object.lower, cpp_object.upper, Bound(cpp_object.left.value), Bound(cpp_object.right.value))

    def as_composite_set(self) -> Interval:
        return Interval(self)

    def __lt__(self, other: Self):
        if self.lower == other.lower:
            return self.upper < other.upper
        return self.lower < other.lower

    def __eq__(self, other: Self):
        return self.lower == other.lower and self.upper == other.upper and self.left == other.left and self.right == other.right

    def is_singleton(self) -> bool:
        """
        :return: True if the interval is a singleton (contains only one value), False otherwise.
        """
        return self.lower == self.upper and self.left == Bound.CLOSED and self.right == Bound.CLOSED

    def contains(self, item: float) -> bool:
        return (self.lower < item < self.upper or (self.lower == item and self.left == Bound.CLOSED) or (
                self.upper == item and self.right == Bound.CLOSED))

    def __hash__(self):
        return hash((self.lower, self.upper, self.left, self.right))

    def non_empty_to_string(self) -> str:
        left_bracket = '[' if self.left == Bound.CLOSED else '('
        right_bracket = ']' if self.right == Bound.CLOSED else ')'
        return f'{left_bracket}{self.lower}, {self.upper}{right_bracket}'

    def __repr__(self):
        return sigma_algebra.AbstractSimpleSet.to_string(self)

    def __str__(self):
        return sigma_algebra.AbstractSimpleSet.to_string(self)

    def to_json(self) -> Dict[str, Any]:
        return {**super().to_json(), 'lower': self.lower, 'upper': self.upper, 'left': self.left.name,
                'right': self.right.name}

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        return cls(data['lower'], data['upper'], Bound[data['left']], Bound[data['right']])

    def center(self) -> float:
        """
        :return: The center point of the interval
        """
        return (self.lower + self.upper) / 2

    def contained_integers(self) -> int:
        """
        :return: Yield integers contained in the interval
        """

        rounded_lower = math.ceil(self.lower)
        if rounded_lower == self.lower and self.left == Bound.OPEN:
            rounded_lower += 1

        rounded_upper = math.floor(self.upper)
        if rounded_upper == self.upper and self.right == Bound.OPEN:
            rounded_upper -= 1

        yield from range(rounded_lower, rounded_upper + 1)

    def __deepcopy__(self):
        return self.__class__(self.lower, self.upper, self.left, self.right)


class Interval(sigma_algebra.AbstractCompositeSet):
    """
    Represents an interval.
    """
    simple_sets: SortedSet[SimpleInterval]

    def __init__(self, *simple_sets):
        """
        Creates a new interval.
        :param simple_sets: The simple intervals that make up the interval.
        """
        super().__init__(*simple_sets)
        self._cpp_object = rl.Interval({simple_set._cpp_object for simple_set in self.simple_sets})

    @classmethod
    def _from_cpp(cls, cpp_object: rl.Interval) -> Self:
        return cls(*[SimpleInterval._from_cpp(cpp_simple_interval) for cpp_simple_interval in cpp_object.simple_sets])

    def new_empty_set(self) -> Self:
        return Interval()

    def complement_if_empty(self) -> Self:
        return Interval([SimpleInterval(float('-inf'), float('inf'), Bound.OPEN, Bound.OPEN)])

    def is_singleton(self):
        """
        :return: True if the interval is a singleton (contains only one value), False otherwise.
        """
        return len(self.simple_sets) == 1 and self.simple_sets[0].is_singleton()

    def contained_integers(self) -> int:
        """
        :return: Yield integers contained in the interval
        """
        for simple_set in self.simple_sets:
            yield from simple_set.contained_integers()


def open(left: float, right: float) -> Interval:
    """
    Creates an open interval.

    :param left: The left bound of the interval.
    :param right: The right bound of the interval.
    :return: The open interval.
    """
    return Interval._from_cpp(rl.open(left, right))


def closed(left: float, right: float) -> Interval:
    """
    Creates a closed interval.

    :param left: The left bound of the interval.
    :param right: The right bound of the interval.
    :return: The closed interval.
    """
    return Interval._from_cpp(rl.closed(left, right))


def open_closed(left: float, right: float) -> Interval:
    """
    Creates an open-closed interval.

    :param left: The left bound of the interval.
    :param right: The right bound of the interval.
    :return: The open-closed interval.
    """
    return Interval._from_cpp(rl.open_closed(left, right))


def closed_open(left: float, right: float) -> Interval:
    """
    Creates a closed-open interval.

    :param left: The left bound of the interval.
    :param right: The right bound of the interval.
    :return: The closed-open interval.
    """
    return Interval._from_cpp(rl.closed_open(left, right))


def singleton(value: float) -> Interval:
    """
    Creates a singleton interval.

    :param value: The value of the interval.
    :return: The singleton interval.
    """
    return Interval._from_cpp(rl.singleton(value))


def reals() -> Interval:
    """
    Creates the set of real numbers.

    :return: The set of real numbers.
    """
    return Interval._from_cpp(rl.reals())



from __future__ import annotations

import itertools
from abc import abstractmethod
from typing import Tuple, Dict, Any

from sortedcontainers import SortedSet
from typing_extensions import Self, Iterable, Optional, TYPE_CHECKING

from random_events.utils import SubclassJSONSerializer

EMPTY_SET_SYMBOL = "∅"


class AbstractSimpleSet(SubclassJSONSerializer):
    """
    Abstract class for simple sets.

    Simple sets are sets that can be represented as a single object.
    """

    @abstractmethod
    def intersection_with(self, other: Self) -> Self:
        """
        Form the intersection of this object with another object.

        :param other: The other SimpleSet
        :return: The intersection of this set with the other set
        """
        raise NotImplementedError

    @abstractmethod
    def complement(self) -> SimpleSetContainer:
        """
        :return: The complement of this set as disjoint set of simple sets.
        """
        raise NotImplementedError

    @abstractmethod
    def is_empty(self) -> bool:
        """
        :return: Rather this set is empty or not.
        """
        raise NotImplementedError

    @abstractmethod
    def contains(self, item) -> bool:
        """
        Check if this set contains an item.
        :param item: The item to check
        :return: Rather if the item is in the set or not
        """
        raise NotImplementedError

    @abstractmethod
    def __hash__(self):
        raise NotImplementedError

    def non_empty_to_string(self) -> str:
        """
        :return: A string representation of this set if it is not empty.
        """
        raise NotImplementedError

    def difference_with(self, other: Self) -> SimpleSetContainer:
        """
        Form the difference of this object with another object.

        :param other: The other SimpleSet
        :return: The difference as disjoint set of simple sets.
        """

        # if the intersection is empty
        intersection = self.intersection_with(other)
        if intersection.is_empty():
            # then the difference is the set itself
            return SortedSet([self])

        # form the complement of the intersection
        complement_of_intersection = intersection.complement()

        # for every element in the complement of the intersection
        result = SortedSet()
        for element in complement_of_intersection:

            # if it intersects with this set
            intersection = element.intersection_with(self)
            if not intersection.is_empty():
                # add the intersection to the result
                result.add(intersection)

        return result

    def to_string(self):
        """
        :return: A string representation of this set.
        """
        if self.is_empty():
            return EMPTY_SET_SYMBOL
        return self.non_empty_to_string()

    def __str__(self):
        return self.to_string()

    @abstractmethod
    def __lt__(self, other):
        raise NotImplementedError

    @abstractmethod
    def as_composite_set(self) -> AbstractCompositeSet:
        """
        Convert this simple set to a composite set.
        :return: The composite set
        """
        raise NotImplementedError

    @abstractmethod
    def __deepcopy__(self):
        raise NotImplementedError


class AbstractCompositeSet(SubclassJSONSerializer):
    """
    Abstract class for composite sets.

    AbstractCompositeSet is a set that is composed of a disjoint union of simple sets.
    """

    simple_sets: SimpleSetContainer

    def __init__(self, *simple_sets):
        self.simple_sets = SortedSet(simple_sets)

    @abstractmethod
    def simplify(self) -> Self:
        """
        Simplify this set into an equivalent, more compact version.

        :return: The simplified set
        """
        raise NotImplementedError

    @abstractmethod
    def new_empty_set(self) -> Self:
        """
        Create a new empty set.

        This method has to be implemented by the subclass and should take over all the relevant attributes to the new
        set.

        :return: A new empty set.
        """
        raise NotImplementedError

    def union_with(self, other: Self) -> Self:
        """
        Form the union of this object with another object.

        :param other: The other set
        :return: The union of this set with the other set
        """
        if other.is_empty():
            return self
        if self.is_empty():
            return other
        result = self.new_empty_set()
        result.simple_sets.update(self.simple_sets)
        result.simple_sets.update(other.simple_sets)
        return result.make_disjoint()

    def __or__(self, other: Self):
        return self.union_with(other)

    def intersection_with_simple_set(self, other: AbstractSimpleSet) -> Self:
        """
        Form the intersection of this object with a simple set.

        :param other: The simple set
        :return: The intersection of this set with the simple set
        """
        result = self.new_empty_set()
        [result.add_simple_set(simple_set.intersection_with(other)) for simple_set in self.simple_sets]
        return result

    def intersection_with_simple_sets(self, other: SimpleSetContainer) -> Self:
        """
        Form the intersection of this object with a set of simple sets.

        :param other: The set of simple sets
        :return: The intersection of this set with the set of simple sets
        """
        result = self.new_empty_set()
        [result.simple_sets.update(self.intersection_with_simple_set(other_simple_set).simple_sets) for other_simple_set
         in other]
        return result

    def intersection_with(self, other: Self) -> Self:
        """
        Form the intersection of this object with another object.
        :param other: The other set
        :return: The intersection of this set with the other set
        """
        return self.intersection_with_simple_sets(other.simple_sets)

    def __and__(self, other):
        return self.intersection_with(other)

    def difference_with_simple_set(self, other: AbstractSimpleSet) -> Self:
        """
        Form the difference with another composite set.
        :param other: The other set
        :return: The difference of this set with the other set
        """
        result = self.new_empty_set()
        [result.simple_sets.update(simple_set.difference_with(other)) for simple_set in self.simple_sets]
        return result.make_disjoint()

    def difference_with_simple_sets(self, other: SimpleSetContainer) -> Self:

        if len(other) == 0:
            return self

        # initialize the result
        result = self.new_empty_set()

        # for every simple set in this set
        for own_simple_set in self.simple_sets:

            # initialize the current difference
            current_difference = self.new_empty_set()
            first_iteration = True

            # for every simple set in the other set
            for other_simple_set in other:

                # form the element wise difference
                difference_with_other_simple_set = own_simple_set.difference_with()

                # if this is the first iteration
                if first_iteration:
                    # just copy the element wise difference
                    current_difference.simple_sets.update(difference_with_other_simple_set)
                    first_iteration = False
                    continue

                # form the intersection of the current difference with the element wise difference
                difference = self.new_empty_set()
                difference.simple_sets.update(difference_with_other_simple_set)
                current_difference = current_difference.intersection_with(difference)

            # add the current difference to the result
            result.simple_sets.update(current_difference.simple_sets)

        return result.make_disjoint()

    def difference_with(self, other: Self) -> Self:
        """
        Form the difference with another composite set.
        :param other: The other set
        :return: The difference of this set with the other set
        """
        return self.difference_with_simple_sets(other.simple_sets)

    def __sub__(self, other):
        return self.difference_with(other)

    def complement(self) -> Self:
        """
        :return: The complement of this set
        """

        if self.is_empty():
            return self.complement_if_empty()

        result = self.new_empty_set()
        result.simple_sets = self.simple_sets[0].complement()
        for simple_set in self.simple_sets[1:]:
            result = result.intersection_with_simple_sets(simple_set.complement())
        return result.make_disjoint()

    @abstractmethod
    def complement_if_empty(self) -> Self:
        """
        :return: The complement of this if it is empty.
        """
        raise NotImplementedError

    def __invert__(self):
        return self.complement()

    def is_empty(self) -> bool:
        """
        Check if this set is empty.
        """
        return len(self.simple_sets) == 0 or all(simple_set.is_empty() for simple_set in self.simple_sets)

    def contains(self, item) -> bool:
        """
        Check if this set contains an item.
        :param item: The item to check
        :return: Rather if the item is in the set or not
        """
        for simple_set in self.simple_sets:
            if simple_set.contains(item):
                return True
        return False

    def __contains__(self, item):
        return self.contains(item)

    def to_string(self):
        """
        :return: A string representation of this set.
        """
        if self.is_empty():
            return EMPTY_SET_SYMBOL
        return " u ".join([simple_set.to_string() for simple_set in self.simple_sets])

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        return self.to_string()

    def is_disjoint(self):
        """
        :return: Rather if the simple sets are disjoint or not.
        """
        for a, b in itertools.combinations(self.simple_sets, 2):
            if not a.intersection_with(b).is_empty():
                return False
        return True

    def split_into_disjoint_and_non_disjoint(self) -> Tuple[Self, Self]:
        """
        Split this composite set into disjoint and non-disjoint parts.

        This method is required for making the composite set disjoint.
        The partitioning is done by removing every other simple set from every simple set.
        The purified simple sets are then disjoint by definition and the pairwise intersections are (potentially) not
        disjoint yet.

        This method requires:
            - the intersection of two simple sets as a simple set
            - the difference_of_a_with_every_b of a simple set (A) and another simple set (B) that is completely contained in A (B ⊆ A).
            The result of that difference_of_a_with_every_b has to be a composite set with only one simple set in it.

        :return: A tuple of the disjoint and non-disjoint set.
        """

        # initialize result for disjoint and non-disjoint sets
        disjoint = self.new_empty_set()
        non_disjoint = self.new_empty_set()

        # for every simple set (a)
        for simple_set_a in self.simple_sets:
            simple_set_a: AbstractSimpleSet

            # initialize the difference of a with every b
            difference_of_a_with_every_b: Optional[AbstractCompositeSet] = self.new_empty_set()
            difference_of_a_with_every_b.add_simple_set(simple_set_a)

            # for every other simple set (b)
            for simple_set_b in self.simple_sets:
                simple_set_b: AbstractSimpleSet

                # skip symmetric iterations
                if simple_set_a == simple_set_b:
                    continue

                # get the intersection of a and b
                intersection_a_b: AbstractSimpleSet = simple_set_a.intersection_with(simple_set_b)

                # if the intersection is not empty, add it to the non-disjoint set
                non_disjoint.add_simple_set(intersection_a_b)

                # get the difference of the simple set with the intersection.
                difference_with_intersection = difference_of_a_with_every_b.difference_with_simple_set(intersection_a_b)

                # if the difference of a with every b is empty
                if len(difference_with_intersection.simple_sets) == 0:
                    # skip the rest of the loop and mark the set for discarding
                    difference_of_a_with_every_b = None
                    break

                # add the disjoint remainder
                difference_of_a_with_every_b = difference_with_intersection

            # if the difference_of_a_with_every_b has become None
            if difference_of_a_with_every_b is None:
                # skip the rest of the loop
                continue

            # append the simple_set_a without every other simple set to the disjoint set
            disjoint.simple_sets.update(difference_of_a_with_every_b.simple_sets)

        return disjoint, non_disjoint

    def make_disjoint(self) -> Self:
        """
        Create an equal composite set that contains a disjoint union of simple sets.

        :return: The disjoint set.
        """

        disjoint, intersection = self.split_into_disjoint_and_non_disjoint()

        # while the intersection is not empty
        while not intersection.is_empty():
            # split the intersection into disjoint and non-disjoint parts
            current_disjoint, intersection = intersection.split_into_disjoint_and_non_disjoint()

            # add the disjoint intersection to the disjoint set
            disjoint.simple_sets.update(current_disjoint.simple_sets)

        return disjoint.simplify()

    def add_simple_set(self, simple_set: AbstractSimpleSet):
        """
        Add a simple set to this composite set if it is not empty.

        :param simple_set: The simple set to add
        """
        if simple_set.is_empty():
            return
        self.simple_sets.add(simple_set)

    def __eq__(self, other: Self):
        return self.simple_sets._list == other.simple_sets._list

    def __hash__(self):
        return hash(tuple(self.simple_sets))

    def __iter__(self):
        return iter(self.simple_sets)

    def __lt__(self, other: Self):
        """
        Compare this set with another set.

        The sets are compared by comparing the simple sets in order.
        If the pair of simple sets are equal, the next pair is compared.
        If all pairs are equal, the set with the least amount of simple sets is considered smaller.

        ..note:: This does not define a total order in the mathematical sense. In the mathematical sense, this defines
            a partial order.

        :param other: The other set
        :return: Rather this set is smaller than the other set
        """
        for a, b in zip(self.simple_sets, other.simple_sets):
            if a == b:
                continue
            else:
                return a < b
        return len(self.simple_sets) < len(other.simple_sets)

    def to_json(self) -> Dict[str, Any]:
        return {**super().to_json(), "simple_sets": [simple_set.to_json() for simple_set in self.simple_sets]}

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        return cls(*[AbstractSimpleSet.from_json(simple_set) for simple_set in data["simple_sets"]])

    def __deepcopy__(self):
        return self.__class__(*[ss.__deepcopy__() for ss in self.simple_sets])


# Type definitions
if TYPE_CHECKING:
    SimpleSetContainer = SortedSet[AbstractSimpleSet]
else:
    SimpleSetContainer = SortedSet



import unittest

import numpy as np

from random_events.polytope import Polytope
import plotly.graph_objects as go

def rotation_matrix(theta: float):
    return np.array([[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]])


class PolytopeTestCase(unittest.TestCase):
    np.random.seed(69)
    points = np.random.uniform(-1, 1, (100, 2))
    points = points @ rotation_matrix(0.3)

    def test_from_2d_points(self):
        polytope = Polytope.from_2d_points(self.points)
        self.assertTrue(polytope.contains(self.points.T).all())

    def test_maximum_inner_box(self):
        polytope = Polytope.from_2d_points(self.points)
        box = polytope.maximum_inner_box()
        # event = box.to_simple_event().as_composite_set()
        # fig = go.Figure()
        # fig.add_trace(go.Scatter(x=self.points[:, 0], y=self.points[:, 1], mode='markers', name='points'))
        # fig.add_traces(event.plot())
        # fig.show()
        self.assertTrue(box <= polytope)

    def test_inner_box_approximation(self):
        polytope = Polytope.from_2d_points(self.points)
        result = polytope.inner_box_approximation(0.1)
        self.assertTrue(result.is_disjoint())
        # fig = go.Figure()
        # fig.add_trace(go.Scatter(x=self.points[:, 0], y=self.points[:, 1], mode='markers', name='points'))
        # fig.add_traces(result.plot())
        # fig.update_layout(result.plotly_layout())
        # fig.show()

    def test_outer_box_approximation(self):
        polytope = Polytope.from_2d_points(self.points)
        result = polytope.outer_box_approximation(0.1)
        self.assertTrue(result.is_disjoint())
        # fig = go.Figure()
        # fig.add_trace(go.Scatter(x=self.points[:, 0], y=self.points[:, 1], mode='markers', name='points'))
        # fig.add_traces(result.plot())
        # fig.update_layout(result.plotly_layout())
        # fig.show()


if __name__ == '__main__':
    unittest.main()



from collections import deque

import numpy as np
import polytope
from ortools.linear_solver import pywraplp
from scipy.spatial import ConvexHull
from typing_extensions import Self, Tuple

from .interval import closed_open
from .product_algebra import Event, SimpleEvent, Continuous


class Polytope(polytope.Polytope):
    """
    Extension of the polytope class from the polytope library.

    This class enables conversion to simple events and provides the inner box and outer box approximation
    from https://cse.lab.imtlucca.it/~bemporad/publications/papers/compgeom-boxes.pdf.
    """

    @classmethod
    def from_polytope(cls, polytope_: polytope.Polytope) -> Self:
        """
        Create a polytope from a polytope object.

        :param polytope_: The polytope object.
        """
        return cls(polytope_.A, polytope_.b)

    @classmethod
    def from_2d_points(cls, points: np.ndarray) -> Self:
        """
        Create a polytope from a set of 2D points, by computing the convex hull of the points and then creating the
        linear inequalities from the convex hull.

        :param points: A numpy array with shape (n, 2) containing the points.
        """

        # create the convexhull
        convex_hull = ConvexHull(points)
        hull_points = np.vstack([points[convex_hull.vertices], points[convex_hull.vertices[0]]])

        # calculate the inequalities
        constraints = []
        for i in range(hull_points.shape[0] - 1):
            p1 = hull_points[i]
            p2 = hull_points[i + 1]
            a = p2[1] - p1[1]
            b = p1[0] - p2[0]
            c = a * p1[0] + b * p1[1]
            constraints.append((a, b, c))
        constraints = np.array(constraints)
        result = cls(constraints[:, :2], constraints[:, 2])
        return result

    def inner_box_approximation(self, minimum_volume: float = 0.1) -> Event:
        """
        Compute an inner box approximation of the polytope.

        Similar to algorithm 5.

        :param minimum_volume: The minimum volume (epsilon) for the approximation.
        If a box is created in the induction with lower volume than epsilon, it will not be split further.

        :return: The inner box approximation of the polytope as a random event.
        """
        # initialize a queue with polytopes that need to be approximated
        working_queue = deque([self])
        resulting_boxes = []

        while working_queue:
            current_polytope = working_queue.popleft()
            inner_box = current_polytope.maximum_inner_box()
            resulting_boxes.append(inner_box)

            # if the inner box is too small, we do not split it further
            if inner_box.volume < minimum_volume:
                continue

            # append the polytope without the inner box to the queue
            diff = polytope.mldivide(current_polytope, inner_box)
            working_queue.extend([self.__class__.from_polytope(p) for p in diff.list_poly])

        return Event(*[box.to_simple_event() for box in resulting_boxes]).make_disjoint()

    def as_box_polytope(self) -> Self:
        """
        :return: The polytope as box polytope.
        """
        lower, upper = self.bounding_box
        return self.from_box([(a[0], b[0]) for a, b in zip(lower, upper)])

    def copy(self):
        return self.from_polytope(self.__copy__())

    def split_on_axis_value(self, axis: int, value: np.ndarray) -> Tuple[Self, Self]:
        """
        Split the polytope on a specific axis and value.

        :param axis: The axis to split on.
        :param value: The value to split on.

        :return: The left and right split of the polytope.
        """
        a_vector = np.zeros((1, self.A.shape[1]))
        a_vector[0, axis] = 1.
        b_vector = value

        # construct left split
        left = self.copy()
        left.A = np.concatenate([left.A, a_vector])
        left.b = np.concatenate([left.b, b_vector])

        # construct right split
        right = self.copy()
        right.A = np.concatenate([right.A, -a_vector])
        right.b = np.concatenate([right.b, -b_vector])

        return left, right

    def outer_box_approximation(self, minimum_volume: float = 0.1) -> Event:
        """
        Compute an outer box approximation of the polytope.
        This implements Algorithm 6 in https://cse.lab.imtlucca.it/~bemporad/publications/papers/compgeom-boxes.pdf

        :param minimum_volume: The minimum volume (epsilon) for the approximation.

        :return: The outer box approximation of the polytope as a random event.
        """
        polytopes_to_split = deque([self])
        resulting_boxes = []

        while polytopes_to_split:
            current_polytope = polytopes_to_split.popleft()
            bounding_box_of_current_polytope = current_polytope.as_box_polytope()

            # if the box is too small, skip
            volume = bounding_box_of_current_polytope.volume
            if volume < minimum_volume or bounding_box_of_current_polytope <= current_polytope:
                resulting_boxes.append(current_polytope)
                continue

            # get the longest side
            lower, upper = current_polytope.bounding_box
            side_lengths = upper - lower
            longest_side = np.argmax(side_lengths)

            # split the box in half along the longest side
            splitting_point = (lower[longest_side] + upper[longest_side]) / 2
            left, right = current_polytope.split_on_axis_value(longest_side, splitting_point)
            polytopes_to_split.extend([left, right])

        return Event(*[box.to_simple_event() for box in resulting_boxes]).make_disjoint()

    def maximum_inner_box(self) -> Self:
        """
        Compute the maximum single inner box approximation of the polytope.

        This implements Algorithm 2 in https://cse.lab.imtlucca.it/~bemporad/publications/papers/compgeom-boxes.pdf

        :return: The maximum inner box of the polytope.
        """

        # calculate bounding box
        minima, maxima = self.bounding_box
        minima = minima.flatten()
        maxima = maxima.flatten()

        solver = pywraplp.Solver.CreateSolver("GLOP")

        # create variables for the dimensions of the inner box approximation (x_0, x_1, ..., x_n)
        dimension_variables = [solver.NumVar(minimum, maximum, f"x_{i}") for i, (minimum, maximum) in
                               enumerate(zip(minima, maxima))]

        # create the scale variable (lambda in the paper)
        scale = solver.NumVar(0, 1, "scale")

        # set the goal to maximize lambda
        solver.Maximize(scale)

        # create the guess for the r vector
        scale_of_box = maxima - minima

        # create the matrix A^+
        a_positive = np.maximum(0, self.A)

        # create the constraints from proposition 2
        for a, a_positive, b in zip(self.A, a_positive, self.b):
            solver.Add(sum(a * dimension_variables) + sum(a_positive * scale_of_box * scale) <= b)

        # solve the problem
        status = solver.Solve()
        assert status == pywraplp.Solver.OPTIMAL, "The solver did not find an optimal solution."

        # calculate the inner box
        box = [[dimension.solution_value(), dimension.solution_value() + scale_of_dimension * scale.solution_value()]
               for dimension, scale_of_dimension in zip(dimension_variables, scale_of_box)]
        return self.__class__.from_box(box)

    def to_simple_event(self) -> SimpleEvent:
        """
        Convert the polytope to a simple event by using its bounding box.
        """
        minima, maxima = self.bounding_box
        minima = minima.flatten()
        maxima = maxima.flatten()
        return SimpleEvent({Continuous(f"x_{i}"): closed_open(minimum, maximum) for i, (minimum, maximum) in
                            enumerate(zip(minima, maxima))})



import cProfile
import unittest
import time
import random
from unittest import skip
import random_events_lib as rl


@skip
class BindingsTestCase(unittest.TestCase):

    def test_intervals_with_bindings_but_in_old_class(self):
        """
        Using bindings, but in the old structure
        """
        from random_events.interval import SimpleInterval, Bound, Interval

        t1 = time.time()
        for i in range(100):
            b = []
            for a in range(50):
                b.append(SimpleInterval(random.randint(0, 100), random.randint(0, 100), Bound.OPEN, Bound.CLOSED))
            z = Interval(*b)
            z.make_disjoint()
        print(f"\nwhen using the bindings within the old classes: {time.time() - t1}\n")

    def test_with_bindings_but_call_bindings_directly(self):
        """
        Only using bindings
        """
        from random_events.interval import Bound

        t1 = time.time()
        for i in range(100):
            b = set()
            for a in range(50):
                b.add(rl.SimpleInterval(random.randint(0, 100), random.randint(0, 100), Bound.OPEN.value, Bound.CLOSED.value))
            z = rl.Interval(b)
            z.make_disjoint()
        print(f"\nwhen calling the bindings directly: {time.time() - t1}\n")

    def test_intervals_old(self):
        """
        No bindings, only python
        """
        from random_events.before_bindings.interval_old import SimpleInterval, Bound, Interval

        t1 = time.time()
        for i in range(100):
            b = []
            for a in range(50):
                b.append(SimpleInterval(random.randint(0, 100), random.randint(0, 100), Bound.OPEN, Bound.CLOSED))
            z = Interval(*b)
            z.make_disjoint()
        print(f"\nwithout bindings: {time.time() - t1}\n")


if __name__ == '__main__':
    cProfile.run('unittest.main()', sort='cumtime')


from abc import abstractmethod

from typing_extensions import Dict, Any, Self


def get_full_class_name(cls):
    """
    Returns the full name of a class, including the module name.

    :param cls: The class.
    :return: The full name of the class
    """
    return cls.__module__ + "." + cls.__name__


def recursive_subclasses(cls):
    """
    :param cls: The class.
    :return: A list of the classes subclasses.
    """
    return cls.__subclasses__() + [g for s in cls.__subclasses__() for g in recursive_subclasses(s)]


class SubclassJSONSerializer:
    """
    Class for automatic (de)serialization of subclasses.
    Classes that inherit from this class can be serialized and deserialized automatically by calling this classes
    'from_json' method.
    """

    def to_json(self) -> Dict[str, Any]:
        return {"type": get_full_class_name(self.__class__)}

    @classmethod
    @abstractmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        """
        Create a variable from a json dict.
        This method is called from the from_json method after the correct subclass is determined and should be
        overwritten by the respective subclass.

        :param data: The json dict
        :return: The deserialized object
        """
        raise NotImplementedError()

    @classmethod
    def from_json(cls, data: Dict[str, Any]) -> Self:
        """
        Create the correct instanceof the subclass from a json dict.

        :param data: The json dict
        :return: The correct instance of the subclass
        """
        for subclass in recursive_subclasses(SubclassJSONSerializer):
            if get_full_class_name(subclass) == data["type"]:
                return subclass._from_json(data)

        raise ValueError("Unknown type {}".format(data["type"]))


import unittest

import numpy as np

from random_events.product_algebra import SimpleEvent
from random_events.variable import *
from random_events.interval import *


str_set = {'a', 'c', 'b'}

class ContinuousTestCase(unittest.TestCase):
    x = Continuous("x")

    def test_creation(self):
        self.assertEqual(self.x.name, "x")
        self.assertEqual(self.x.domain, reals())

    def test_to_json(self):
        x_ = Variable.from_json(self.x.to_json())
        self.assertEqual(self.x, x_)


class IntegerTestCase(unittest.TestCase):

    def test_creation(self):
        x = Integer("x")
        self.assertEqual(x.name, "x")
        self.assertEqual(x.domain, reals())


class SymbolicTestCase(unittest.TestCase):

    def test_creation(self):
        a = SetElement("a", str_set)
        b = SetElement("b", str_set)
        c = SetElement("c", str_set)
        x = Symbolic("x", Set(a, b, c))
        self.assertEqual(x.name, "x")
        self.assertEqual(x.domain, Set(a, b, c))

    def test_to_json(self):
        a = SetElement("a", str_set)
        b = SetElement("b", str_set)
        c = SetElement("c", str_set)
        x = Symbolic("x", Set(a, b, c))
        x_ = Variable.from_json(x.to_json())
        self.assertEqual(x, x_)

class Continuous2(Continuous):
    mean: int

    def __init__(self, name, mean):
        super().__init__(name)
        self.mean = mean

class InheritanceTestCase(unittest.TestCase):

    def test_conversion(self):
        v1 = Continuous2("david", 2)
        event = SimpleEvent({v1: open_closed(-np.inf, 0)}).as_composite_set()
        event2 = event.complement()
        v2 = event2.all_variables[0]
        self.assertIsInstance(v2, Continuous2)


if __name__ == '__main__':
    unittest.main()



from abc import abstractmethod
import random_events_lib as rl

from typing_extensions import Self, Type, Dict, Any, Union, Optional

from .interval import Interval, reals
from .set import Set, SetElement
from .sigma_algebra import AbstractCompositeSet, EMPTY_SET_SYMBOL
from .utils import SubclassJSONSerializer


class Variable(SubclassJSONSerializer):
    """
    Parent class for all random variables.
    """
    name: str
    domain: AbstractCompositeSet

    def __init__(self, name: str, domain: AbstractCompositeSet):
        """
        Construct a new random variable.
        :param name: The name of the variable.
        :param domain: The domain of the variable.
        """
        self.name = name
        self.domain = domain

    def __lt__(self, other: Self) -> bool:
        """
        Returns True if self < other, False otherwise.
        """
        return self.name < other.name

    def __gt__(self, other: Self) -> bool:
        """
        Returns True if self > other, False otherwise.
        """
        return self.name > other.name

    def __hash__(self) -> int:
        return self.name.__hash__()

    def __eq__(self, other):
        return self.name == other.name

    def __str__(self):
        return f"{self.__class__.__name__}({self.name}, {self.domain})"

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name})"

    def to_json(self) -> Dict[str, Any]:
        return {
            **super().to_json(),
            "name": self.name,
            "domain": self.domain.to_json()
        }

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        return cls(data["name"], AbstractCompositeSet.from_json(data["domain"]))

    @property
    @abstractmethod
    def is_numeric(self):
        """
        :return: Rather, this variable has a numeric domain or not.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def _from_cpp(cls, cpp_object):
        if type(cpp_object) == rl.Symbolic:
            return Symbolic._from_cpp(cpp_object)
        elif type(cpp_object) == rl.Continuous:
            return Continuous._from_cpp(cpp_object)
        elif type(cpp_object) == rl.Integer:
            return Integer._from_cpp(cpp_object)


class Continuous(Variable):
    """
    Class for continuous random variables.

    The domain of a continuous variable is the real line.
    """

    def __init__(self, name: str, domain: AbstractCompositeSet = reals()):
        """
        Construct a continuous variable.
        :param name: The name.
        :param domain: The domain of the variable.
        """
        super().__init__(name, reals())

        self._cpp_object = rl.Continuous(self.name)

    @property
    def is_numeric(self):
        return True

    @classmethod
    def _from_cpp(cls, cpp_object):
        return cls(cpp_object.name)


class Symbolic(Variable):
    """
    Class for unordered, finite, discrete random variables.

    The domain of a symbolic variable is a set of values from an enumeration.
    """
    domain: Set

    def __init__(self, name: Optional[str] = None, domain: Union[Type[SetElement], Set] = None):
        """
        Construct a symbolic variable.
        :param name: The name.
        :param domain: The enum class that lists all elements of the domain.
        """

        if domain is None:
            raise ValueError("The domain of a symbolic variable must be specified.")

        if isinstance(domain, set):
            self.domain = Set(*[SetElement(value, domain) for value in domain if value != EMPTY_SET_SYMBOL])
        else:
            self.domain = domain

        super().__init__(name, self.domain)

        self._cpp_object = rl.Symbolic(self.name, self.domain._cpp_object)

    def domain_type(self) -> Type[SetElement]:
        return self.domain.simple_sets[0].all_elements

    @property
    def is_numeric(self):
        return False

    @classmethod
    def _from_cpp(cls, cpp_object):
        return cls(cpp_object.name, Set._from_cpp(cpp_object.get_domain()))


class Integer(Variable):
    """
    Class for ordered, discrete random variables.

    The domain of an integer variable is the number line.
    """

    def __init__(self, name: str, domain: AbstractCompositeSet = reals()):
        """
        Construct an integer variable.
        :param name: The name.
        :param domain: The domain of the variable.
        """
        super().__init__(name, reals())

        self._cpp_object = rl.Integer(self.name)

    @property
    def is_numeric(self):
        return True

    @classmethod
    def _from_cpp(cls, cpp_object):
        return cls(cpp_object.name)


from abc import abstractmethod

from typing_extensions import Self, Type, Dict, Any, Union, Optional

from random_events.before_bindings.interval_old import Interval, reals

from random_events.before_bindings.sigma_algebra_old import *
from random_events.utils import SubclassJSONSerializer


class Variable(SubclassJSONSerializer):
    name: str
    domain: AbstractCompositeSet

    def __init__(self, name: str, domain: AbstractCompositeSet):
        self.name = name
        self.domain = domain

    def __lt__(self, other: Self) -> bool:
        """
        Returns True if self < other, False otherwise.
        """
        return self.name < other.name

    def __gt__(self, other: Self) -> bool:
        """
        Returns True if self > other, False otherwise.
        """
        return self.name > other.name

    def __hash__(self) -> int:
        return self.name.__hash__()

    def __eq__(self, other):
        return self.name == other.name

    def __str__(self):
        return f"{self.__class__.__name__}({self.name}, {self.domain})"

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name})"

    def to_json(self) -> Dict[str, Any]:
        return {
            **super().to_json(),
            "name": self.name,
            "domain": self.domain.to_json()
        }

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        return cls(data["name"], AbstractCompositeSet.from_json(data["domain"]))

    @property
    @abstractmethod
    def is_numeric(self):
        """
        :return: Rather, this variable has a numeric domain or not.
        """
        raise NotImplementedError


class Continuous(Variable):
    """
    Class for continuous random variables.

    The domain of a continuous variable is the real line.
    """
    domain: Interval

    def __init__(self, name: str, domain=None):
        super().__init__(name, reals())

    @property
    def is_numeric(self):
        return True


# class Symbolic(Variable):
#     """
#     Class for unordered, finite, discrete random variables.
#
#     The domain of a symbolic variable is a set of values from an enumeration.
#     """
#     domain: Set
#
#     def __init__(self, name: Optional[str] = None, domain: Union[Type[SetElement], Set] = None):
#         """
#         Construct a symbolic variable.
#         :param name: The name.
#         :param domain: The enum class that lists all elements of the domain.
#         """
#
#         if domain is None:
#             raise ValueError("The domain of a symbolic variable must be specified.")
#
#         if isinstance(domain, Set):
#             domain = domain.simple_sets[0].__class__
#
#         assert isinstance(domain, type) and issubclass(domain, SetElement), ("The domain must be a subclass of "
#                                                                              "SetElement.")
#         if name is None:
#             name = domain.__name__
#
#         super().__init__(name, Set(*[value for value in domain if value != domain.EMPTY_SET]))
#
#     def domain_type(self) -> Type[SetElement]:
#         return self.domain.simple_sets[0].all_elements
#
#     @property
#     def is_numeric(self):
#         return False
#

class Integer(Variable):
    """
    Class for ordered, discrete random variables.

    The domain of an integer variable is the number line.
    """
    domain: Interval

    def __init__(self, name: str, domain=None):
        super().__init__(name, reals())

    @property
    def is_numeric(self):
        return True


from __future__ import annotations

import enum
import math
from dataclasses import dataclass
from typing import Dict, Any

from sortedcontainers import SortedSet
from typing_extensions import Self

from random_events.before_bindings.sigma_algebra_old import *


class Bound(enum.Enum):
    """
    Enumerates the possible bounds for an interval.
    """

    CLOSED = 0
    """
    Represents a closed bound, i. e. the element is included from the interval.
    """

    OPEN = 1
    """
    Represents an open bound, i. e. the element is excluded in the interval.
    """

    def invert(self):
        return Bound.CLOSED if self == Bound.OPEN else Bound.OPEN

    def intersect(self, other: Self) -> Self:
        """
        Intersect with another border

        :param other: The other border
        :return: The intersection of the two borders
        """
        return Bound.OPEN if self == Bound.OPEN or other == Bound.OPEN else Bound.CLOSED


@dataclass
class SimpleInterval(AbstractSimpleSet):
    """
    Represents a simple interval.
    """

    lower: float = 0
    """
    The lower bound of the interval.
    """

    upper: float = 0
    """
    The upper bound of the interval.
    """

    left: Bound = Bound.OPEN
    """
    The bound type of the lower bound.
    """

    right: Bound = Bound.OPEN
    """
    The bound type of the upper bound.
    """

    def as_composite_set(self) -> Interval:
        return Interval(self)

    def __lt__(self, other: Self):
        if self.lower == other.lower:
            return self.upper < other.upper
        return self.lower < other.lower

    def is_empty(self) -> bool:
        return self.lower > self.upper or (
                self.lower == self.upper and (self.left == Bound.OPEN or self.right == Bound.OPEN))

    def is_singleton(self) -> bool:
        """
        :return: True if the interval is a singleton (contains only one value), False otherwise.
        """
        return self.lower == self.upper and self.left == Bound.CLOSED and self.right == Bound.CLOSED

    def intersection_with(self, other: Self) -> Self:

        # create new limits for the intersection
        new_lower = max(self.lower, other.lower)
        new_upper = min(self.upper, other.upper)

        # if the new limits are not valid, return an empty interval
        if new_lower > new_upper:
            return SimpleInterval()

        # create the new left bound
        if self.lower == other.lower:
            new_left = self.left.intersect(other.left)
        else:
            new_left = self.left if self.lower > other.lower else other.left

        # create the new right bound
        if self.upper == other.upper:
            new_right = self.right.intersect(other.right)
        else:
            new_right = self.right if self.upper < other.upper else other.right

        return SimpleInterval(new_lower, new_upper, new_left, new_right)

    def complement(self) -> SortedSet[Self]:

        # if the interval is empty
        if self.is_empty():
            # return the real line
            return SortedSet([SimpleInterval(float('-inf'), float('inf'), Bound.OPEN, Bound.OPEN)])

        # initialize the result
        result = SortedSet()

        # if this is the real line
        if self.lower == float('-inf') and self.upper == float('inf'):
            # return the empty set
            return result

        # if the lower bound is not negative infinity
        if self.lower > float('-inf'):
            # add the interval from minus infinity to the lower bound
            result.add(SimpleInterval(float('-inf'), self.lower, Bound.OPEN, self.left.invert()))

        # if the upper bound is not positive infinity
        if self.upper < float('inf'):
            # add the interval from the upper bound to infinity
            result.add(SimpleInterval(self.upper, float('inf'), self.right.invert(), Bound.OPEN))

        return result

    def contains(self, item: float) -> bool:
        return (self.lower < item < self.upper or (self.lower == item and self.left == Bound.CLOSED) or (
                self.upper == item and self.right == Bound.CLOSED))

    def __hash__(self):
        return hash((self.lower, self.upper, self.left, self.right))

    def non_empty_to_string(self) -> str:
        left_bracket = '[' if self.left == Bound.CLOSED else '('
        right_bracket = ']' if self.right == Bound.CLOSED else ')'
        return f'{left_bracket}{self.lower}, {self.upper}{right_bracket}'

    def __repr__(self):
        return AbstractSimpleSet.to_string(self)

    def __str__(self):
        return AbstractSimpleSet.to_string(self)

    def to_json(self) -> Dict[str, Any]:
        return {**super().to_json(), 'lower': self.lower, 'upper': self.upper, 'left': self.left.name,
                'right': self.right.name}

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        return cls(data['lower'], data['upper'], Bound[data['left']], Bound[data['right']])

    def center(self) -> float:
        """
        :return: The center point of the interval
        """
        return (self.lower + self.upper) / 2

    def contained_integers(self) -> int:
        """
        :return: Yield integers contained in the interval
        """

        rounded_lower = math.ceil(self.lower)
        if rounded_lower == self.lower and self.left == Bound.OPEN:
            rounded_lower += 1

        rounded_upper = math.floor(self.upper)
        if rounded_upper == self.upper and self.right == Bound.OPEN:
            rounded_upper -= 1

        yield from range(rounded_lower, rounded_upper + 1)

    def __deepcopy__(self):
        return self.__class__(self.lower, self.upper, self.left, self.right)


class Interval(AbstractCompositeSet):
    simple_sets: SortedSet[SimpleInterval]

    def simplify(self) -> Self:

        # if the set is empty, return it
        if self.is_empty():
            return self

        # initialize the result
        result = self.simple_sets[0].as_composite_set()

        # iterate over the simple sets
        for current_simple_interval in self.simple_sets[1:]:

            # get the last element in the result
            last_simple_interval = result.simple_sets[-1]

            # if the borders are connected
            if (last_simple_interval.upper > current_simple_interval.lower or (
                    last_simple_interval.upper == current_simple_interval.lower and not (
                    last_simple_interval.right == Bound.OPEN and current_simple_interval.left == Bound.OPEN))):

                # extend the upper bound of the last element
                last_simple_interval.upper = current_simple_interval.upper
                last_simple_interval.right = current_simple_interval.right
            else:

                # add the current element to the result
                result.simple_sets.add(current_simple_interval)

        return result

    def new_empty_set(self) -> Self:
        return Interval()

    def complement_if_empty(self) -> Self:
        return Interval([SimpleInterval(float('-inf'), float('inf'), Bound.OPEN, Bound.OPEN)])

    def is_singleton(self):
        """
        :return: True if the interval is a singleton (contains only one value), False otherwise.
        """
        return len(self.simple_sets) == 1 and self.simple_sets[0].is_singleton()

    def contained_integers(self) -> int:
        """
        :return: Yield integers contained in the interval
        """
        for simple_set in self.simple_sets:
            yield from simple_set.contained_integers()


def open(left: float, right: float) -> Interval:
    """
    Creates an open interval.

    :param left: The left bound of the interval.
    :param right: The right bound of the interval.
    :return: The open interval.
    """
    return SimpleInterval(left, right, Bound.OPEN, Bound.OPEN).as_composite_set()


def closed(left: float, right: float) -> Interval:
    """
    Creates a closed interval.

    :param left: The left bound of the interval.
    :param right: The right bound of the interval.
    :return: The closed interval.
    """
    return SimpleInterval(left, right, Bound.CLOSED, Bound.CLOSED).as_composite_set()


def open_closed(left: float, right: float) -> Interval:
    """
    Creates an open-closed interval.

    :param left: The left bound of the interval.
    :param right: The right bound of the interval.
    :return: The open-closed interval.
    """
    return SimpleInterval(left, right, Bound.OPEN, Bound.CLOSED).as_composite_set()


def closed_open(left: float, right: float) -> Interval:
    """
    Creates a closed-open interval.

    :param left: The left bound of the interval.
    :param right: The right bound of the interval.
    :return: The closed-open interval.
    """
    return SimpleInterval(left, right, Bound.CLOSED, Bound.OPEN).as_composite_set()


def singleton(value: float) -> Interval:
    """
    Creates a singleton interval.

    :param value: The value of the interval.
    :return: The singleton interval.
    """
    return SimpleInterval(value, value, Bound.CLOSED, Bound.CLOSED).as_composite_set()


def reals() -> Interval:
    """
    Creates the set of real numbers.

    :return: The set of real numbers.
    """
    return SimpleInterval(float('-inf'), float('inf'), Bound.OPEN, Bound.OPEN).as_composite_set()



from __future__ import annotations

import itertools

import numpy as np
from random_events_lib import SimpleInterval
from sortedcontainers import SortedDict, SortedKeysView, SortedValuesView
from typing_extensions import List, TYPE_CHECKING
import plotly.graph_objects as go

from random_events.before_bindings.sigma_algebra_old import *
from random_events.before_bindings.variable_old import *



# Type definitions
if TYPE_CHECKING:
    VariableMapSuperClassType = SortedDict[Variable, Any]
else:
    VariableMapSuperClassType = SortedDict


class VariableMap(VariableMapSuperClassType):
    """
    A map of variables to values.

    Accessing a variable by name is also supported.
    """

    @property
    def variables(self) -> SortedKeysView[Variable]:
        return self.keys()

    def variable_of(self, name: str) -> Variable:
        """
        Get the variable with the given name.
        :param name: The variable's name
        :return: The variable itself
        """
        variable = [variable for variable in self.variables if variable.name == name]
        if len(variable) == 0:
            raise KeyError(f"Variable {name} not found in event {self}")
        return variable[0]

    def __getitem__(self, item: Union[str, Variable]):
        if isinstance(item, str):
            item = self.variable_of(item)
        return super().__getitem__(item)

    def __setitem__(self, key: Union[str, Variable], value: Any):
        if isinstance(key, str):
            key = self.variable_of(key)

        if not isinstance(key, Variable):
            raise TypeError(f"Key must be a Variable if not already present, got {type(key)} instead.")

        super().__setitem__(key, value)

    def __copy__(self):
        return self.__class__({variable: value for variable, value in self.items()})


class SimpleEvent(AbstractSimpleSet, VariableMap):
    """
    A simple event is a set of assignments of variables to values.

    A simple event is logically equivalent to a conjunction of assignments.
    """

    def __init__(self, *args, **kwargs):
        VariableMap.__init__(self, *args, **kwargs)
        for key, value in self.items():
            self[key] = value

    def as_composite_set(self) -> Event:
        return Event(self)

    @property
    def assignments(self) -> SortedValuesView:
        return self.values()

    def intersection_with(self, other: Self) -> Self:
        variables = self.keys() | other.keys()
        result = SimpleEvent()
        for variable in variables:
            if variable in self and variable in other:
                result[variable] = self[variable].intersection_with(other[variable])
            elif variable in self:
                result[variable] = self[variable]
            else:
                result[variable] = other[variable]

        return result

    def complement(self) -> SimpleSetContainer:

        # initialize result
        result = SortedSet()

        # initialize variables where the complement has already been computed
        processed_variables = []

        # for every key, value pair
        for variable, assignment in self.items():

            # initialize the current complement
            current_complement = SimpleEvent()

            # set the current variable to its complement
            current_complement[variable] = assignment.complement()

            # for every other variable
            for other_variable in self.variables:

                # skip this iteration if the other variable is the same as the current one
                if other_variable == variable:
                    continue

                # if it has been processed, set copy its assignment from this
                if other_variable in processed_variables:
                    current_complement[other_variable] = self[other_variable]

                # otherwise, set it to its domain (set of all values)
                else:
                    current_complement[other_variable] = other_variable.domain

            # memorize the processed variables
            processed_variables.append(variable)

            # if the current complement is not empty, add it to the result
            if not current_complement.is_empty():
                result.add(current_complement)

        return result

    def is_empty(self) -> bool:
        if len(self.keys()) == 0:
            return True

        for assignment in self.values():
            if assignment.is_empty():
                return True

        return False

    def contains(self, item: Tuple) -> bool:
        for assignment, value in zip(self.assignments, item):
            if not assignment.contains(value):
                return False
        return True

    def __hash__(self):
        return hash(tuple(self.items()))

    def __setitem__(self, key: Variable, value: Union[AbstractSimpleSet, AbstractCompositeSet]):
        if isinstance(value, AbstractSimpleSet):
            super().__setitem__(key, value.as_composite_set())
        elif isinstance(value, AbstractCompositeSet):
            super().__setitem__(key, value)
        else:
            raise TypeError(f"Value must be a SimpleSet or CompositeSet, got {type(value)} instead.")

    def __lt__(self, other: Self):
        if len(self.variables) < len(other.variables):
            return True
        for variable in self.variables:
            if variable not in other:
                return True
            if self[variable] == other[variable]:
                continue
            else:
                return self[variable] < other[variable]

    def marginal(self, variables: VariableSet) -> SimpleEvent:
        """
        Create the marginal event, that only contains the variables given..

        :param variables: The variables to contain in the marginal event
        :return: The marginal event
        """
        result = self.__class__()
        for variable in variables:
            result[variable] = self[variable]
        return result

    def non_empty_to_string(self) -> str:
        return "{" + ", ".join(f"{variable.name} = {assignment}" for variable, assignment in self.items()) + "}"

    def variables_to_json(self) -> List:
        return [variable.to_json() for variable in self.keys()]

    def assignments_to_json(self) -> List:
        return [assignment.to_json() for assignment in self.values()]

    def to_json(self) -> Dict[str, Any]:
        return {**super().to_json(),
                "variables": self.variables_to_json(),
                "assignments": self.assignments_to_json()}

    def to_json_assignments_only(self) -> Dict[str, Any]:
        return {**super().to_json(),
                "assignments": self.assignments_to_json()}

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        variables = [Variable.from_json(variable) for variable in data["variables"]]
        assignments = [AbstractCompositeSet.from_json(assignment) for assignment in data["assignments"]]
        return cls({variable: assignment for variable, assignment in zip(variables, assignments)})

    @classmethod
    def from_json_given_variables(cls, data: Dict[str, Any], variables: List[Variable]) -> Self:
        assignments = [AbstractCompositeSet.from_json(assignment) for assignment in data["assignments"]]
        return cls({variable: assignment for variable, assignment in zip(variables, assignments)})

    def plot(self) -> Union[List[go.Scatter], List[go.Mesh3d]]:
        """
        Plot the event.
        """
        assert all(isinstance(variable, Continuous) for variable in self.keys()), \
            "Plotting is only supported for events that consist of only continuous variables."
        if len(self.keys()) == 1:
            return self.plot_1d()
        if len(self.keys()) == 2:
            return self.plot_2d()
        elif len(self.keys()) == 3:
            return self.plot_3d()
        else:
            raise NotImplementedError("Plotting is only supported for two and three dimensional events")

    def plot_1d(self) -> List[go.Scatter]:
        """
        Plot the event in 1D.
        """
        xs = []
        ys = []

        interval: Interval = list(self.values())[0]
        for simple_interval in interval.simple_sets:
            simple_interval: SimpleInterval
            xs.extend([simple_interval.lower, simple_interval.upper, None])
            ys.extend([0, 0, None])

        return [go.Scatter(x=xs, y=ys, mode="lines", name="Event", fill="toself")]

    def plot_2d(self) -> List[go.Scatter]:
        """
        Plot the event in 2D.
        """

        # form cartesian product of all intervals
        intervals = [value.simple_sets for value in self.values()]
        interval_combinations = list(itertools.product(*intervals))

        xs = []
        ys = []

        # for every atomic interval
        for interval_combination in interval_combinations:

            # plot a rectangle
            points = np.asarray(list(itertools.product(*[[axis.lower, axis.upper] for axis in interval_combination])))
            y_points = points[:, 1]
            y_points[len(y_points) // 2:] = y_points[len(y_points) // 2:][::-1]
            xs.extend(points[:, 0].tolist() + [points[0, 0], None])
            ys.extend(y_points.tolist()+ [y_points[0], None])

        return [go.Scatter(x=xs, y=ys, mode="lines", name="Event", fill="toself")]

    def plot_3d(self) -> List[go.Mesh3d]:
        """
        Plot the event in 3D.
        """

        # form cartesian product of all intervals
        intervals = [value.simple_sets for _, value in sorted(self.items())]
        simple_events = list(itertools.product(*intervals))
        traces = []

        # shortcut for the dimensions
        x, y, z = 0, 1, 2

        # for every atomic interval
        for simple_event in simple_events:

            # Create a 3D mesh trace for the rectangle
            traces.append(go.Mesh3d(
                # 8 vertices of a cube
                x=[simple_event[x].lower, simple_event[x].lower, simple_event[x].upper, simple_event[x].upper,
                   simple_event[x].lower, simple_event[x].lower, simple_event[x].upper, simple_event[x].upper],
                y=[simple_event[y].lower, simple_event[y].upper, simple_event[y].upper, simple_event[y].lower,
                   simple_event[y].lower, simple_event[y].upper, simple_event[y].upper, simple_event[y].lower],
                z=[simple_event[z].lower, simple_event[z].lower, simple_event[z].lower, simple_event[z].lower,
                   simple_event[z].upper, simple_event[z].upper, simple_event[z].upper, simple_event[z].upper],
                # i, j and k give the vertices of triangles
                i=[7, 0, 0, 0, 4, 4, 6, 6, 4, 0, 3, 2],
                j=[3, 4, 1, 2, 5, 6, 5, 2, 0, 1, 6, 3],
                k=[0, 7, 2, 3, 6, 7, 1, 1, 5, 5, 7, 6],
                flatshading=True
            ))
        return traces

    def plotly_layout(self) -> Dict:
        """
        Create a layout for the plotly plot.
        """
        if len(self.variables) == 1:
            result = {"xaxis_title": self.variables[0].name}
        elif len(self.variables) == 2:
            result = {"xaxis_title": self.variables[0].name,
                      "yaxis_title": self.variables[1].name}
        elif len(self.variables) == 3:
            result = dict(scene=dict(
                xaxis_title=self.variables[0].name,
                yaxis_title=self.variables[1].name,
                zaxis_title=self.variables[2].name)
            )
        else:
            raise NotImplementedError("Plotting is only supported for two and three dimensional events")

        return result

    def fill_missing_variables(self, variables: VariableSet):
        """
        Fill this with the variables that are not in self but in `variables`.
        The variables are mapped to their domain.

        :param variables: The variables to fill the event with
        """
        for variable in variables:
            if variable not in self:
                self[variable] = variable.domain

    def __deepcopy__(self):
        return self.__class__({variable: assignment.__deepcopy__() for variable, assignment in self.items()})


class Event(AbstractCompositeSet):
    """
    An event is a disjoint set of simple events.

    Every simple event added to this event that is missing variables that any other event in this event has, will be
    extended with the missing variable. The missing variables are mapped to their domain.

    """

    simple_sets: SimpleEventContainer

    def __init__(self, *simple_sets):
        super().__init__(*simple_sets)
        self.fill_missing_variables()

    @property
    def all_variables(self) -> VariableSet:
        result = SortedSet()
        return result.union(*[SortedSet(simple_set.variables) for simple_set in self.simple_sets])

    def fill_missing_variables(self):
        """
        Fill all simple sets with the missing variables.
        """
        all_variables = self.all_variables
        for simple_set in self.simple_sets:
            simple_set.fill_missing_variables(all_variables)

    def simplify(self) -> Self:
        simplified, changed = self.simplify_once()
        while changed:
            simplified, changed = simplified.simplify_once()
        return simplified

    def simplify_once(self) -> Tuple[Self, bool]:
        """
        Simplify the event once. This simplification is not guaranteed to as simple as possible.

        :return: The simplified event and a boolean indicating whether the event has changed or not.
        """

        for event_a, event_b in itertools.combinations(self.simple_sets, 2):
            different_variables = SortedSet()

            # get all events where these two events differ
            for variable in event_a.variables:
                if event_a[variable] != event_b[variable]:
                    different_variables.add(variable)

                # if the pair of simple events mismatches in more than one dimension it cannot be simplified
                if len(different_variables) > 1:
                    break

            # if the pair of simple events mismatches in more than one dimension skip it
            if len(different_variables) > 1:
                continue

            # get the dimension where the two events differ
            different_variable = different_variables[0]

            # initialize the simplified event
            simplified_event = SimpleEvent()

            # for every variable
            for variable in event_a.variables:

                # if the variable is the one where the two events differ
                if variable == different_variable:
                    # set it to the union of the two events
                    simplified_event[variable] = event_a[variable].union_with(event_b[variable])

                # if the variable has the same assignment
                else:
                    # copy to the simplified event
                    simplified_event[variable] = event_a[variable]

            # create a new event with the simplified event and all other events
            result = Event(
                *([simplified_event] + [event for event in self.simple_sets if event != event_a and event != event_b]))
            return result, True

        # if nothing happened, return the original event and False
        return self, False

    def new_empty_set(self) -> Self:
        return Event()

    def complement_if_empty(self) -> Self:
        raise NotImplementedError("Complement of an empty Event is not yet supported.")

    def marginal(self, variables: VariableSet) -> Event:
        """
        Create the marginal event, that only contains the variables given..

        :param variables: The variables to contain in the marginal event
        :return: The marginal event
        """
        result = self.__class__()
        for simple_set in self.simple_sets:
            result.add_simple_set(simple_set.marginal(variables))
        return result.make_disjoint()

    def bounding_box(self) -> SimpleEvent:
        """
        Compute the bounding box of the event.
        The bounding box is the smallest simple event that contains this event. It is computed by taking the union
        of all simple events variable wise.

        :return: The bounding box as a simple event
        """
        result = SimpleEvent()
        for variable in self.all_variables:
            for simple_set in self.simple_sets:
                if variable not in result:
                    result[variable] = simple_set[variable].__deepcopy__()
                else:
                    result[variable] = result[variable].union_with(simple_set[variable].__deepcopy__())
        return result

    def plot(self, color="#636EFA") -> Union[List[go.Scatter], List[go.Mesh3d]]:
        """
        Plot the complex event.

        :param color: The color to use for this event
        """
        traces = []
        show_legend = True
        for index, event in enumerate(self.simple_sets):
            event_traces = event.plot()
            for event_trace in event_traces:
                if len(event.keys()) == 2:
                    event_trace.update(name="Event", legendgroup=id(self), showlegend=show_legend,
                                       line=dict(color=color))
                if len(event.keys()) == 3:
                    event_trace.update(name="Event", legendgroup=id(self), showlegend=show_legend, color=color)
                show_legend = False
                traces.append(event_trace)
        return traces

    def plotly_layout(self) -> Dict:
        """
        Create a layout for the plotly plot.
        """
        return self.simple_sets[0].plotly_layout()

    def add_simple_set(self, simple_set: AbstractSimpleSet):
        """
        Add a simple set to this event.

        :param simple_set: The simple set to add
        """
        super().add_simple_set(simple_set)
        self.fill_missing_variables()

    def to_json(self) -> Dict[str, Any]:
        variables = [variable.to_json() for variable in self.all_variables]
        simple_sets = [simple_set.to_json_assignments_only() for simple_set in self.simple_sets]
        return {**SubclassJSONSerializer.to_json(self),
                "variables": variables, "simple_sets": simple_sets}

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        variables = [Variable.from_json(variable) for variable in data["variables"]]
        simple_sets = [SimpleEvent.from_json_given_variables(simple_set, variables) for simple_set in data["simple_sets"]]
        return cls(*simple_sets)


# Type definitions
if TYPE_CHECKING:
    SimpleEventContainer = SortedSet[SimpleEvent]
    EventContainer = SortedSet[Event]
    VariableSet = SortedSet[Variable]
else:
    SimpleEventContainer = SortedSet
    EventContainer = SortedSet
    VariableSet = SortedSet



from random_events.set import SetElement, Set, EMPTY_SET_SYMBOL
import enum
import unittest

from random_events.sigma_algebra import AbstractSimpleSet


str_set = {'a', 'c', 'b'}
int_set = {1, 2, 3}
float_set = {1.001, 2.3845, 3.4345345}

EMPTY_SET = SetElement(EMPTY_SET_SYMBOL, set())

class SetElementTestCase(unittest.TestCase):

    def test_intersection_with(self):
        a = SetElement('a', str_set)
        b = SetElement('b', str_set)

        intersection_a_b = a.intersection_with(b)
        self.assertEqual(intersection_a_b, EMPTY_SET)
        self.assertEqual(a.intersection_with(a), a)

    def test_complement(self):
        a = SetElement('a', str_set)
        b = SetElement('b', str_set)
        c = SetElement('c', str_set)
        # a1 = SetElement(1, int_set)
        # a2 = SetElement(2, int_set)
        # a3 = SetElement(3, int_set)
        complement_a = a.complement()
        complement_b = b.complement()
        complement_c = c.complement()
        # complement_a1 = a1.complement()
        self.assertEqual(complement_a, {b, c})
        self.assertEqual(complement_b, {a, c})
        self.assertEqual(complement_c, {a, b})
        # self.assertEqual(complement_a1, {a2, a3})

    def test_contains(self):
        a = SetElement('a', str_set)
        b = SetElement('b', str_set)
        c = SetElement('c', str_set)
        self.assertTrue(a.contains(a))
        self.assertFalse(a.contains(b))
        self.assertFalse(a.contains(c))

    def test_to_json(self):
        a = SetElement('a', str_set)
        b = AbstractSimpleSet.from_json(a.to_json())
        self.assertEqual(a, b)


class SetTestCase(unittest.TestCase):

    def test_simplify(self):
        a = SetElement('a', str_set)
        b = SetElement('b', str_set)
        c = SetElement('c', str_set)
        s = Set(a, b, c, c)
        self.assertEqual(3, len(s.simple_sets))
        ss = Set(a, b, c)
        self.assertEqual(ss, s.simplify())

    def test_difference(self):
        a = SetElement('a', str_set)
        b = SetElement('b', str_set)
        s = Set(a, b)
        s_ = Set(a)
        self.assertEqual(s.difference_with(s_), Set(b))

    def test_complement(self):
        a = SetElement('a', str_set)
        b = SetElement('b', str_set)
        c = SetElement('c', str_set)
        s = Set(a, b)
        self.assertEqual(s.complement(), Set(c))

    def test_to_json(self):
        a = SetElement('a', str_set)
        b = SetElement('b', str_set)
        s = Set(a, b)
        s_ = AbstractSimpleSet.from_json(s.to_json())
        self.assertEqual(s, s_)


class SetTypeTestCase(unittest.TestCase):
    def test_int(self):
        a = SetElement(1, int_set)
        b = SetElement(2, int_set)
        c = SetElement(3, int_set)

        s = Set(a, b, c)
        s2 = Set(a, b)
        inter = s.intersection_with(s2)
        self.assertEqual(inter, s2)

    def test_str(self):
        a = SetElement('a', str_set)
        b = SetElement('b', str_set)
        c = SetElement('c', str_set)

        s = Set(a, b, c)
        s2 = Set(a, b)
        inter = s.intersection_with(s2)
        self.assertEqual(inter, s2)

    def test_float(self):
        a = SetElement(1.001, float_set)
        b = SetElement(2.3845, float_set)
        c = SetElement(3.4345345, float_set)

        s = Set(a, b, c)
        s2 = Set(a, b)
        inter = s.intersection_with(s2)
        self.assertEqual(inter, s2)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from abc import abstractmethod
from typing import Tuple, Dict, Any
import random_events_lib as rl
from sortedcontainers import SortedSet
from typing_extensions import Self, Iterable, Optional, TYPE_CHECKING, Set
from .utils import SubclassJSONSerializer

EMPTY_SET_SYMBOL = "∅"


class AbstractSimpleSet(SubclassJSONSerializer):
    """
    Abstract class for simple sets.

    Simple sets are sets that can be represented as a single object.
    """

    @abstractmethod
    def __init__(self):
        """
        Create a new simple set.
        """
        self._cpp_object = rl.AbstractSimpleSet

    @abstractmethod
    def _from_cpp(self, cpp_object):
        """
        Create a new instance of this class from a C++ object.
        """
        raise NotImplementedError

    def intersection_with(self, other: Self) -> Self:
        """
        Form the intersection of this object with another object.

        :param other: The other SimpleSet
        :return: The intersection of this set with the other set
        """
        return self._from_cpp(self._cpp_object.intersection_with(other._cpp_object))

    def complement(self) -> SimpleSetContainer:
        """
        :return: The complement of this set as disjoint set of simple sets.
        """
        return SortedSet({self._from_cpp(cpp_simple_set) for cpp_simple_set in self._cpp_object.complement()})

    def is_empty(self) -> bool:
        """
        :return: Rather this set is empty or not.
        """
        return self._cpp_object.is_empty()

    @abstractmethod
    def contains(self, item) -> bool:
        """
        Check if this set contains an item.
        :param item: The item to check
        :return: Rather if the item is in the set or not
        """
        raise NotImplementedError

    @abstractmethod
    def __hash__(self):
        raise NotImplementedError

    def non_empty_to_string(self) -> str:
        """
        :return: A string representation of this set if it is not empty.
        """
        raise NotImplementedError

    def difference_with(self, other: Self) -> SimpleSetContainer:
        """
        Form the difference of this object with another object.

        :param other: The other SimpleSet
        :return: The difference as a disjoint set of simple sets.
        """
        return self._from_cpp(self._cpp_object.difference_with(other._cpp_object))

    def to_string(self):
        """
        :return: A string representation of this set.
        """
        if self.is_empty():
            return EMPTY_SET_SYMBOL
        return self.non_empty_to_string()

    def __str__(self):
        return self.to_string()

    @abstractmethod
    def __lt__(self, other):
        raise NotImplementedError

    @abstractmethod
    def as_composite_set(self) -> AbstractCompositeSet:
        """
        Convert this simple set to a composite set.
        :return: The composite set
        """
        raise NotImplementedError

    @abstractmethod
    def __deepcopy__(self):
        raise NotImplementedError


class AbstractCompositeSet(SubclassJSONSerializer):
    """
    Abstract class for composite sets.

    AbstractCompositeSet is a set that is composed of a disjoint union of simple sets.
    """

    simple_sets: SimpleSetContainer

    def __init__(self, *simple_sets):
        self.simple_sets = SortedSet(simple_sets)

        self._cpp_object = rl.AbstractCompositeSet

    @classmethod
    @abstractmethod
    def _from_cpp(cls, cpp_object):
        raise NotImplementedError

    @abstractmethod
    def new_empty_set(self) -> Self:
        """
        Create a new empty set.

        This method has to be implemented by the subclass and should take over all the relevant attributes to the new
        set.

        :return: A new empty set.
        """
        raise NotImplementedError

    def union_with(self, other: Self) -> Self:
        """
        Form the union of this object with another object.

        :param other: The other set
        :return: The union of this set with the other set
        """
        return self._from_cpp(self._cpp_object.union_with(other._cpp_object))

    def __or__(self, other: Self):
        return self.union_with(other)

    def intersection_with_simple_set(self, other: AbstractSimpleSet) -> Self:
        """
        Form the intersection of this object with a simple set.

        :param other: The simple set
        :return: The intersection of this set with the simple set
        """
        return self._from_cpp(self._cpp_object.intersection_with_simple_set(other._cpp_object))

    def intersection_with_simple_sets(self, other: SimpleSetContainer) -> Self:
        """
        Form the intersection of this object with a set of simple sets.

        :param other: The set of simple sets
        :return: The intersection of this set with the set of simple sets
        """
        result = self.new_empty_set()
        [result.simple_sets.update(self.intersection_with_simple_set(other_simple_set).simple_sets) for other_simple_set
         in other]
        return result

    def intersection_with(self, other: Self) -> Self:
        """
        Form the intersection of this object with another object.
        :param other: The other set
        :return: The intersection of this set with the other set
        """
        return self._from_cpp(self._cpp_object.intersection_with(other._cpp_object))

    def __and__(self, other):
        return self.intersection_with(other)

    def difference_with_simple_set(self, other: AbstractSimpleSet) -> Self:
        """
        Form the difference with another composite set.
        :param other: The other set
        :return: The difference of this set with the other set
        """
        return self._from_cpp(self._cpp_object.difference_with(other._cpp_object))

    def difference_with_simple_sets(self, other: SimpleSetContainer) -> Self:
        """
        Form the difference with a set of simple sets.
        :param other: The set of simple sets
        :return: The difference of this set with the other set
        """
        result = self.new_empty_set()
        [result.simple_sets.update(self.difference_with_simple_set(other_simple_set).simple_sets) for other_simple_set
         in other]
        return result

    def difference_with(self, other: Self) -> Self:
        """
        Form the difference with another composite set.
        :param other: The other set
        :return: The difference of this set with the other set
        """
        return self._from_cpp(self._cpp_object.difference_with(other._cpp_object))

    def __sub__(self, other):
        return self.difference_with(other)

    def complement(self) -> Self:
        """
        :return: The complement of this set
        """
        return self._from_cpp(self._cpp_object.complement())

    @abstractmethod
    def complement_if_empty(self) -> Self:
        """
        :return: The complement of this if it is empty.
        """
        raise NotImplementedError

    def __invert__(self):
        return self.complement()

    def is_empty(self) -> bool:
        """
        Check if this set is empty.
        """
        return self._cpp_object.is_empty()

    def contains(self, item) -> bool:
        """
        Check if this set contains an item.
        :param item: The item to check
        :return: Rather if the item is in the set or not
        """
        for simple_set in self.simple_sets:
            if simple_set.contains(item):
                return True
        return False

    def __contains__(self, item):
        return self.contains(item)

    def to_string(self):
        """
        :return: A string representation of this set.
        """
        if self.is_empty():
            return EMPTY_SET_SYMBOL
        return " u ".join([simple_set.to_string() for simple_set in self.simple_sets])

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        return self.to_string()

    def is_disjoint(self):
        """
        :return: Rather if the simple sets are disjoint or not.
        """
        return self._cpp_object.is_disjoint()

    def make_disjoint(self) -> Self:
        """
        Create an equal composite set that contains a disjoint union of simple sets.

        :return: The disjoint set.
        """
        return self._from_cpp(self._cpp_object.make_disjoint())

    def add_simple_set(self, simple_set: AbstractSimpleSet):
        """
        Add a simple set to this composite set if it is not empty.

        :param simple_set: The simple set to add
        """
        if simple_set.is_empty():
            return
        self.simple_sets.add(simple_set)
        self._cpp_object.add_new_simple_set(simple_set._cpp_object)

    def simplify(self) -> Self:
        """
        Simplify this set into an equivalent, more compact version.

        :return: The simplified set
        """
        return self._from_cpp(self._cpp_object.simplify())

    def __eq__(self, other: Self):
        return self.simple_sets._list == other.simple_sets._list

    def __hash__(self):
        return hash(tuple(self.simple_sets))

    def __iter__(self):
        return iter(self.simple_sets)

    def __lt__(self, other: Self):
        """
        Compare this set with another set.

        The sets are compared by comparing the simple sets in order.
        If the pair of simple sets are equal, the next pair is compared.
        If all pairs are equal, the set with the least amount of simple sets is considered smaller.

        ..note:: This does not define a total order in the mathematical sense. In the mathematical sense, this defines
            a partial order.

        :param other: The other set
        :return: Rather this set is smaller than the other set
        """
        for a, b in zip(self.simple_sets, other.simple_sets):
            if a == b:
                continue
            else:
                return a < b
        return len(self.simple_sets) < len(other.simple_sets)

    def to_json(self) -> Dict[str, Any]:
        return {**super().to_json(), "simple_sets": [simple_set.to_json() for simple_set in self.simple_sets]}

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        return cls(*[AbstractSimpleSet.from_json(simple_set) for simple_set in data["simple_sets"]])

    def __deepcopy__(self):
        return self.__class__(*[ss.__deepcopy__() for ss in self.simple_sets])


# Type definitions
if TYPE_CHECKING:
    SimpleSetContainer = SortedSet[AbstractSimpleSet]
else:
    SimpleSetContainer = SortedSet



import unittest

from random_events.interval import *
from random_events.sigma_algebra import AbstractSimpleSet


class SimpleIntervalTestCase(unittest.TestCase):

    def test_intersection(self):
        a = SimpleInterval(0, 1)
        b = SimpleInterval(0.5, 2)
        c = SimpleInterval(0.5, 0.75, Bound.OPEN, Bound.CLOSED)

        intersection_a_b = a.intersection_with(b)
        intersection_a_b_ = SimpleInterval(0.5, 1, Bound.OPEN, Bound.OPEN)
        self.assertEqual(intersection_a_b, intersection_a_b_)

        intersection_a_c = a.intersection_with(c)
        self.assertEqual(intersection_a_c, c)

    def test_is_empty(self):
        a = SimpleInterval()
        b = SimpleInterval(3, 1)
        c = SimpleInterval(0, 1)
        self.assertTrue(a.is_empty())
        self.assertTrue(b.is_empty())
        self.assertFalse(c.is_empty())

    def test_complement(self):
        a = SimpleInterval()
        complement_a = a.complement()
        self.assertEqual(complement_a, SortedSet([SimpleInterval(-float('inf'), float('inf'), Bound.OPEN, Bound.OPEN)]))
        b = SimpleInterval(0, 1)
        complement_b = b.complement()
        self.assertEqual(complement_b, SortedSet([SimpleInterval(-float('inf'), 0, Bound.OPEN, Bound.CLOSED),
                                                  SimpleInterval(1, float('inf'), Bound.CLOSED, Bound.OPEN)]))

    def test_contains(self):
        a = SimpleInterval(0, 1)
        self.assertFalse(a.contains(0))
        self.assertTrue(a.contains(0.5))
        self.assertFalse(a.contains(1))
        self.assertFalse(a.contains(-1))
        self.assertFalse(a.contains(1.1))

    def test_to_json(self):
        a = SimpleInterval(0, 1)
        b = AbstractSimpleSet.from_json(a.to_json())
        self.assertIsInstance(b, SimpleInterval)
        self.assertEqual(a, b)


class IntervalTestCase(unittest.TestCase):

    def test_simplify(self):
        a = SimpleInterval(0, 1)
        b = SimpleInterval(0.5, 1.5)
        c = SimpleInterval(1.5, 2, Bound.CLOSED)
        d = SimpleInterval(3, 4)
        a_b = Interval(d, a, b, c)
        a_b_simplified = a_b.simplify()
        a_b_simplified_ = Interval(SimpleInterval(0, 2), SimpleInterval(3, 4))
        self.assertEqual(a_b_simplified, a_b_simplified_)

    def test_intersection_with_self(self):
        a = SimpleInterval(0, 1)
        b = SimpleInterval(0.5, 1.5)
        c = SimpleInterval(1.5, 2, Bound.CLOSED)
        d = SimpleInterval(3, 4)
        a_d = Interval(a, d)
        b_c = Interval(b, c)

        intersection_a_d_b_c = a_d.intersection_with(b_c)
        intersection_expected = Interval(SimpleInterval(0.5, 1))
        self.assertEqual(intersection_a_d_b_c, intersection_expected)

    def test_intersection_with_simple_set(self):
        a = SimpleInterval(0, 1)
        b = SimpleInterval(0.5, 1.5)
        d = SimpleInterval(3, 4)
        a_d = Interval(a, d)

        intersection_a_d_b_c = a_d.intersection_with_simple_set(b)
        intersection_expected = Interval(SimpleInterval(0.5, 1))
        self.assertEqual(intersection_a_d_b_c, intersection_expected)

    def test_intersection_with_set_of_simple_sets(self):
        a = SimpleInterval(0, 1)
        b = SimpleInterval(0.5, 1.5)
        c = SimpleInterval(1.5, 2, Bound.CLOSED)
        d = SimpleInterval(3, 4)
        a_d = Interval(a, d)

        intersection_a_d_b_c = a_d.intersection_with_simple_sets({b, c})
        intersection_expected = Interval(SimpleInterval(0.5, 1))
        self.assertEqual(intersection_a_d_b_c, intersection_expected)

    def test_complement(self):
        a = SimpleInterval(0, 1)
        d = SimpleInterval(3, 4)
        a_d = Interval(a, d)

        complement_a_d = a_d.complement()
        complement_a_d_ = Interval(SimpleInterval(-float('inf'), 0, Bound.OPEN, Bound.CLOSED),
                                   SimpleInterval(1, 3, Bound.CLOSED, Bound.CLOSED),
                                   SimpleInterval(4, float('inf'), Bound.CLOSED, Bound.OPEN))
        self.assertEqual(complement_a_d, complement_a_d_)

    def test_make_disjoint(self):
        a = SimpleInterval(0, 1, Bound.CLOSED, Bound.CLOSED)
        b = SimpleInterval(0.5, 1.5, Bound.CLOSED, Bound.CLOSED)
        c = SimpleInterval(1.5, 2, Bound.CLOSED, Bound.CLOSED)
        d = SimpleInterval(2, 3, Bound.CLOSED, Bound.CLOSED)
        a_b_c_d = Interval(a, b, c, d)

        disjoint = a_b_c_d.make_disjoint()

        a_b_c_d_expected = Interval(SimpleInterval(0, 3, Bound.CLOSED, Bound.CLOSED))
        self.assertEqual(disjoint, a_b_c_d_expected)

    def test_union(self):
        a = SimpleInterval(0, 1)
        b = SimpleInterval(0.5, 1.5)
        c = SimpleInterval(1.5, 2, Bound.CLOSED)
        d = SimpleInterval(3, 4)
        a_d = Interval(a, d)
        b_c = Interval(b, c)

        union_a_d_b_c = a_d.union_with(b_c)
        union_a_d_b_c_ = Interval(SimpleInterval(0, 2), SimpleInterval(3, 4))
        self.assertEqual(union_a_d_b_c, union_a_d_b_c_)
        self.assertTrue(union_a_d_b_c.is_disjoint())

    def test_to_json(self):
        a = SimpleInterval(0, 1)
        b = Interval(a)
        c = AbstractSimpleSet.from_json(b.to_json())
        self.assertIsInstance(c, Interval)
        self.assertEqual(b, c)

    def test_partial_order(self):
        a = open(2, 4) | open(5, 6)
        b = open(3, 4) | open(4.5, 5.5)
        self.assertTrue(a < b)
        self.assertFalse(b < a)

    def test_contained_integers(self):
        a = open(2, 4) | closed_open(4.5, 6)
        self.assertEqual(list(a.contained_integers()), [3, 5])


if __name__ == '__main__':
    unittest.main()
